import warnings
from typing import Callable, Any

import polars as pl
import pandas as pd

from data_fingerprint.src.models import RowDifference, DataReport


def _convert_parameters_to_polars(*args, **kwargs) -> tuple[tuple, dict]:
    """
    Convert pandas DataFrames to polars DataFrames.

    Raises:
        UserWarning: If a pandas DataFrame is found.

    Args:
        *args: The arguments to convert.
        **kwargs: The keyword arguments to convert.

    Returns:
        tuple[tuple, dict]: The converted arguments and keyword arguments.
    """

    def transform_pandas_to_polars(arg):
        warnings.warn(
            "Trasnforming pandas DataFrames to polars DataFrames. "
            "There may be some data types changes. "
            "Please transform DataFrames to polars before analyzing and "
            "find out the differences.",
            UserWarning,
        )
        return pl.from_pandas(arg)

    arg: list[pl.DataFrame] = [
        transform_pandas_to_polars(arg) if isinstance(arg, pd.DataFrame) else arg
        for arg in args
    ]
    kwa: dict[str, pl.DataFrame] = {
        key: (
            transform_pandas_to_polars(value)
            if isinstance(value, pd.DataFrame)
            else value
        )
        for key, value in kwargs.items()
    }
    return arg, kwa


def convert_to_polars(func: Callable) -> Callable:
    """
    Decorator to convert pandas DataFrames to polars DataFrames.
    If the DataFrame is already a polars DataFrame, it will be returned as is.

    This decorator is useful when you want to use a function that expects polars DataFrames,
    but you have pandas DataFrames.
    It will convert the pandas DataFrames to polars DataFrames before calling the function.
    It will also warn the user that the DataFrames are being converted.
    It is recommended to convert the DataFrames to polars before analyzing and finding the differences.
    This will avoid any data type changes that may occur during the conversion.
    The conversion is done using the `pl.from_pandas` function.

    Args:
        func (Callable): The function to decorate.

    Returns:
        Callable: The decorated function.
    """

    def wrapper(*args, **kwargs) -> pl.DataFrame:
        arg, kwa = _convert_parameters_to_polars(*args, **kwargs)
        return func(*arg, **kwa)

    return wrapper


def convert_row_differences_to_pandas(
    row_differences: list[RowDifference],
) -> pl.DataFrame:
    """
    Convert a list of :class:`data_compare.src.models.RowDifference` objects to a pandas DataFrame.

    Args:
        row_differences (list[:class:`data_compare.src.models.RowDifference`]): The list of :class:`data_compare.src.models.RowDifference` objects.

    Returns:
        pl.DataFrame: The resulting DataFrame.
        If the list is empty, returns an empty DataFrame.
        The DataFrame will have the same columns as the :class:`data_compare.src.models.RowDifference` objects.
        The DataFrame will have an additional column "source" with the source of the row.
    """
    row_differences_list: list[pl.DataFrame] = []

    for row_diff in row_differences:
        df: pl.DataFrame = pl.DataFrame(row_diff.row)
        df = df.with_columns(pl.lit(row_diff.source).alias("source"))
        row_differences_list.append(df)

    if len(row_differences_list) == 0:
        return pl.DataFrame()

    return pl.concat(row_differences_list)


def get_dataframe(data_report: DataReport) -> pl.DataFrame:
    """
    Convert a :class:`data_compare.src.models.DataReport` object to a pandas DataFrame.

    This function will convert the row differences to a pandas DataFrame.
    The DataFrame will have only columns that are comparable between the two dataframes (*more information in :class:`data_compare.src.models.DataReport`*).

    .. note::
       This function will return an empty DataFrame if there are no row differences.

    Args:
        data_report (:class:`data_compare.src.models.DataReport`): The :class:`data_compare.src.models.DataReport` object.

    Returns:
        pl.DataFrame: The resulting DataFrame.
    """
    gathered_rows: list[pl.DataFrame] = []

    for rd in data_report.row_differences:
        tmp_rows: pl.DataFrame = pl.DataFrame(rd.row)
        if isinstance(rd, RowDifference):
            tmp_rows = tmp_rows.with_columns(pl.lit(rd.source).alias("source"))
            gathered_rows.append(tmp_rows)
            continue

        tmp_rows: pl.DataFrame = pl.DataFrame(rd.row_with_source)
        gathered_rows.append(tmp_rows)

    if len(gathered_rows) == 0:
        return pl.DataFrame()

    return pl.concat(gathered_rows, how="vertical_relaxed")



from enum import Enum


class ColumnNameDifferenceType(str, Enum):
    """
    Enum for column name difference types.
    """

    MISSING: str = "MISSING"
    """When this column is missing from the first source."""

    EXTRA: str = "EXTRA"
    """When this column is extra in the first source (**not** in the second source)."""

    DIFFERENT_TYPE: str = "DIFFERENT_TYPE"
    """When this column is present in both sources but has a different data type."""


class ColumnDataTypeDifferenceType(str, Enum):
    """
    Enum for column data type difference types.
    This is used when the column names are the same but the data types are different.
    """

    DIFFERENT_TYPE: str = "DIFFERENT_TYPE"
    """When this column is present in both sources but has a different data type."""

    DIFFERENT_TIMEZONE: str = "DIFFERENT_TIMEZONE"
    """When this column is present in both sources but has a different timezone."""

    DIFFERENT_TIME_PRECISION: str = "DIFFERENT_TIME_PRECISION"
    """When this column is present in both sources but has a different time precision."""


class RowDifferenceType(str, Enum):
    """
    Enum for row difference types.
    """

    MISSING_ROW: str = "MISSING_ROW"
    """When this row is present in `source` row but not in the other source."""



from typing import Union, Optional

import polars as pl

from data_fingerprint.src.models import (
    ColumnDifference,
    RowDifference,
    RowGroupDifference,
    DataReport,
)
from data_fingerprint.src.utils import (
    convert_to_polars,
    convert_row_differences_to_pandas,
)
from data_fingerprint.src.checkers import check_inputs
from data_fingerprint.src.difference_types import (
    ColumnNameDifferenceType,
    ColumnDataTypeDifferenceType,
    RowDifferenceType,
)


@convert_to_polars
@check_inputs
def get_column_name_differences(
    df0: pl.DataFrame, df1: pl.DataFrame, df0_name: str, df1_name: str
) -> tuple[list[str], list[ColumnDifference]]:
    """
    Get the differences in column names between two `polars.DataFrame`.
    The referent DataFrame is `df0`. The differences are returned as a list of :class:`data_compare.src.models.ColumnDifference` objects.

    Example:
        ```python
        import polars as pl
        from data_compare.src.comparator import get_column_name_differences
        from data_compare.src.models import ColumnDifference
        df0 = pl.DataFrame({"a": [1, 2], "b": [3, 4]})
        df1 = pl.DataFrame({"a": [1, 2], "c": [3, 4]})
        df0_name = "df0"
        df1_name = "df1"
        same_columns, different_columns = get_column_name_differences(df0, df1, df0_name, df1_name)
        print(same_columns)
        print(different_columns)
        ```
    Result:
        ```
        {'a'}

        [ColumnDifference(source='df0', column_name='c',
        difference_type=<ColumnNameDifferenceType.MISSING: 'MISSING'>, more_information=None),
        ColumnDifference(source='df0', column_name='b',
        difference_type=<ColumnNameDifferenceType.EXTRA: 'EXTRA'>, more_information=None)]
        ```

    Args:
        df0 (pl.DataFrame): The first DataFrame.
        df1 (pl.DataFrame): The second DataFrame.
        df0_name (str): The name of the first DataFrame.
        df1_name (str): The name of the second DataFrame.

    Returns:
        list[str]: A list of column names that are the same in both DataFrames.

        list[ColumnDifference]: A list of :class:`data_compare.src.models.ColumnDifference` objects representing the differences in column names.
    """
    column_names_0 = set(df0.columns)
    column_names_1 = set(df1.columns)

    # Get the differences in column names
    # Extra columns are columns in df0 that are not in df1
    extra_columns = column_names_0 - column_names_1
    # Missing columns are columns in df1 that are not in df0
    missing_columns = column_names_1 - column_names_0

    # get the same column names
    same_columns = column_names_0 & column_names_1

    column_differences: list[ColumnDifference] = []
    for missing_col in missing_columns:
        column_differences.append(
            ColumnDifference(
                source=df0_name,
                column_name=missing_col,
                difference_type=ColumnNameDifferenceType.MISSING,
            )
        )
    for extra_col in extra_columns:
        column_differences.append(
            ColumnDifference(
                source=df0_name,
                column_name=extra_col,
                difference_type=ColumnNameDifferenceType.EXTRA,
            )
        )
    return same_columns, column_differences


@convert_to_polars
@check_inputs
def get_column_dtype_differences(
    df0: pl.DataFrame, df1: pl.DataFrame, df0_name: str, df1_name: str
) -> tuple[list[str], list[ColumnDifference]]:
    """
    Get the differences in column types between two `polars.DataFrame` objects.

    This function only checks differences in types on columns that are present in both dataframes
    (same column names, see :func:`get_column_name_differences` for more details).

    Currently the function is capable of cathing the following differences:
    - Different column types (e.g. `pl.Int64` vs `pl.Float64`)
    - Different time zones of same column type (e.g. `pl.Datetime(time_zone="UTC")` vs `pl.Datetime(time_zone="Europe/Berlin")`)
    - Different time precisions of same column type (e.g. `pl.Datetime(time_unit="ms")` vs `pl.Datetime(time_unit="us")`)

    Example:
        ```python
        import polars as pl
        from data_compare.src.comparator import get_column_dtype_differences

        df0 = pl.DataFrame(
            {
                "a": [1, 2, 3],
                "b": [4, 5, 6],
            }
        )
        df1 = pl.DataFrame(
            {
                "a": [1, 2, 3],
                "b": [4.0, 5.0, 6.0],
            }
        )

        same_columns, column_differences = get_column_dtype_differences(df0, df1, "df0", "df1")
        print(same_columns)
        print(column_differences)
        ```
        Output:
        ```
        ['a']

        [ColumnDifference(source='df0', column_name='b',
        difference_type=<ColumnDataTypeDifferenceType.DIFFERENT_TYPE: 'DIFFERENT_TYPE'>,
        more_information={'df0': Int64, 'df1': Float64})]
        ```


    .. note::
        **One thing to remember is that timezone of `pl.Datetime` column is defined by first value in the column.**

    Args:
        df0 (pl.DataFrame): The first dataframe.
        df1 (pl.DataFrame): The second dataframe.
        df0_name (str): The name of the first dataframe.
        df1_name (str): The name of the second dataframe.

    Returns:
        list[str]: The names of the columns that have the same type in both dataframes.

        list[:class:`data_compare.src.models.ColumnDifference`]: The differences in column types between the two dataframes.

    """
    df0_dtypes: dict[str, str] = {
        column_name: type(dtype) for column_name, dtype in zip(df0.columns, df0.dtypes)
    }
    df1_dtypes: dict[str, str] = {
        column_name: type(dtype) for column_name, dtype in zip(df1.columns, df1.dtypes)
    }

    same_columns, column_differences = get_column_name_differences(
        df0, df1, df0_name, df1_name
    )

    same_columns_after_dtpe_check: list[str] = []
    same_columns_after_dtpe_check.extend(same_columns)

    for same_col in same_columns:
        if df0_dtypes[same_col] != df1_dtypes[same_col]:
            column_differences.append(
                ColumnDifference(
                    source=df0_name,
                    column_name=same_col,
                    difference_type=ColumnDataTypeDifferenceType.DIFFERENT_TYPE,
                    more_information={
                        df0_name: df0_dtypes[same_col],
                        df1_name: df1_dtypes[same_col],
                    },
                )
            )
            same_columns_after_dtpe_check.remove(same_col)
            continue

        # check if the timezone is the same
        # in polars.DataFrame first element of the column is default timezone for the column
        if df0_dtypes[same_col] == pl.Datetime:
            if df0[same_col][0].tzinfo != df1[same_col][0].tzinfo:
                column_differences.append(
                    ColumnDifference(
                        source=df0_name,
                        column_name=same_col,
                        difference_type=ColumnDataTypeDifferenceType.DIFFERENT_TIMEZONE,
                        more_information={
                            df0_name: df0[same_col][0].tzinfo,
                            df1_name: df1[same_col][0].tzinfo,
                        },
                    )
                )
                same_columns_after_dtpe_check.remove(same_col)
                continue

            # check if the precision of time is the same
            if df0[same_col].dtype.time_unit != df1[same_col].dtype.time_unit:
                column_differences.append(
                    ColumnDifference(
                        source=df0_name,
                        column_name=same_col,
                        difference_type=ColumnDataTypeDifferenceType.DIFFERENT_TIME_PRECISION,
                        more_information={
                            df0_name: df0[same_col].dtype.time_unit,
                            df1_name: df1[same_col].dtype.time_unit,
                        },
                    )
                )
                same_columns_after_dtpe_check.remove(same_col)
                continue

    return same_columns_after_dtpe_check, column_differences


@convert_to_polars
@check_inputs
def get_row_differences(
    df0: pl.DataFrame,
    df1: pl.DataFrame,
    df0_name: str,
    df1_name: str,
) -> tuple[list[str], list[ColumnDifference], list[RowDifference]]:
    """
    Get the row differences between two dataframes, meaning find the rows that are in one dataframe but not in the other **or they differ**.


    This function compares only the rows that have the same columns and data types (:func:`get_column_dtype_differences`).


    Good thing is that this function also detects the duplicate rows, and also rows that are duplicated by a different number of times in each dataframe.

    Example:
        ```python
        import polars as pl
        from data_compare.src.comparator import get_row_differences

        df0 = pl.DataFrame(
            {
                "a": [1, 2, 3, 4],
                "b": [4, 5, 6, 7],
            }
        )
        df1 = pl.DataFrame(
            {
                "a": [1, 2, 3],
                "b": [4.0, 5, 6.0],
            }
        )

        same_columns, column_differences, row_differences = get_row_differences(df0, df1, "df0", "df1")
        print(same_columns)
        print(column_differences)
        print(row_differences)
        ```
        Output:
        ```
        ['a']

        [ColumnDifference(source='df0', column_name='b',
        difference_type=<ColumnDataTypeDifferenceType.DIFFERENT_TYPE: 'DIFFERENT_TYPE'>,
        more_information={'df0': Int64, 'df1': Float64})]

        [RowDifference(source='df0',
        row={'a': [4]}, number_of_occurrences=1,
        difference_type=<RowDifferenceType.MISSING_ROW: 'MISSING_ROW'>,
        more_information=None)]
        ```

    Args:
        df0 (pl.DataFrame): The first dataframe.
        df1 (pl.DataFrame): The second dataframe.
        df0_name (str): The name of the first dataframe.
        df1_name (str): The name of the second dataframe.

    Returns:
       list[str]: The columns that are the same

       list[:class:`data_compare.src.models.ColumnDifference`]: The column differences (for more info see :func:`get_column_dtype_differences`)

       list[:class:`data_compare.src.models.RowDifference`]: The row differences.


    """
    same_columns, column_differences = get_column_dtype_differences(
        df0, df1, df0_name, df1_name
    )

    df0_subset: pl.DataFrame = df0.select(same_columns)
    df1_subset: pl.DataFrame = df1.select(same_columns)

    df0_subset = df0_subset.with_columns(df0_subset.hash_rows().alias("hash"))
    df1_subset = df1_subset.with_columns(df1_subset.hash_rows().alias("hash"))

    df0_subset = df0_subset.with_columns(pl.lit(df0_name).alias("source"))
    df1_subset = df1_subset.with_columns(pl.lit(df1_name).alias("source"))

    row_differences: list[RowDifference] = []

    differences_hash_df0: set[str] = set(df0_subset["hash"]).difference(
        set(df1_subset["hash"])
    )
    differences_hash_df1: set[str] = set(df1_subset["hash"]).difference(
        set(df0_subset["hash"])
    )

    for difference_hash in differences_hash_df0:
        difference_row: pl.DataFrame = df0_subset.filter(
            pl.col("hash") == difference_hash
        )

        diff: RowDifference = RowDifference(
            source=df0_name,
            row=difference_row.select(sorted(difference_row.columns))
            .drop(["hash", "source"])
            .sort("*")
            .to_dict(as_series=False),
            number_of_occurrences=difference_row.shape[0],
            difference_type=RowDifferenceType.MISSING_ROW,
        )
        row_differences.append(diff)

    for difference_hash in differences_hash_df1:
        difference_row: pl.DataFrame = df1_subset.filter(
            pl.col("hash") == difference_hash
        )

        diff: RowDifference = RowDifference(
            source=df1_name,
            row=difference_row.select(sorted(difference_row.columns))
            .drop(["hash", "source"])
            .sort("*")
            .to_dict(as_series=False),
            number_of_occurrences=difference_row.shape[0],
            difference_type=RowDifferenceType.MISSING_ROW,
        )
        row_differences.append(diff)

    # look for duplicates that are in both dataframes
    # but not the same number of times
    duplicates_df0 = df0_subset.filter(pl.col("hash").is_duplicated())
    duplicates_df1 = df1_subset.filter(pl.col("hash").is_duplicated())

    duplicates_df0_hashes: set[str] = set(duplicates_df0["hash"]) & set(
        df1_subset["hash"]
    )
    duplicates_df1_hashes: set[str] = set(duplicates_df1["hash"]) & set(
        df0_subset["hash"]
    )

    same_hashes: set[str] = duplicates_df0_hashes.union(duplicates_df1_hashes)
    for same_hash in same_hashes:
        duplicates_df0_count = df0_subset.filter(pl.col("hash") == same_hash).shape[0]
        duplicates_df1_count = df1_subset.filter(pl.col("hash") == same_hash).shape[0]

        if duplicates_df0_count == duplicates_df1_count:
            continue

        if duplicates_df0_count != duplicates_df1_count:
            if duplicates_df0_count > duplicates_df1_count:
                diff: RowDifference = RowDifference(
                    source=df0_name,
                    row=df0_subset.filter(pl.col("hash") == same_hash)
                    .select(sorted(df0_subset.columns))
                    .drop(["hash", "source"])
                    .sort("*")
                    .head(duplicates_df0_count - duplicates_df1_count)
                    .to_dict(as_series=False),
                    number_of_occurrences=duplicates_df0_count - duplicates_df1_count,
                    difference_type=RowDifferenceType.MISSING_ROW,
                )
                row_differences.append(diff)
                continue

            diff: RowDifference = RowDifference(
                source=df1_name,
                row=df1_subset.filter(pl.col("hash") == same_hash)
                .select(sorted(df1_subset.columns))
                .drop(["hash", "source"])
                .sort("*")
                .head(duplicates_df1_count - duplicates_df0_count)
                .to_dict(as_series=False),
                number_of_occurrences=duplicates_df1_count - duplicates_df0_count,
                difference_type=RowDifferenceType.MISSING_ROW,
            )
            row_differences.append(diff)

    return same_columns, column_differences, row_differences


def compare_group_column_by_column(
    data: pl.DataFrame, grouping_columns: list[str]
) -> list[Union[RowDifference, RowGroupDifference]]:
    """
    Compares the rows of a dataframe (**already**) grouped by the grouping columns.


    The main idea of this function is to recieve a dataframe that is grouped
    by the grouping columns and has a "source" column that indicates the source of the row.
    The function will then compare the rows of the dataframe and return the differences between the rows of the different sources.

    This function will return a list of :class:`data_compare.src.models.RowDifference` or :class:`data_compare.src.models.RowGroupDifference` objects:
    - The :class:`data_compare.src.models.RowDifference` object will be returned when there are only one source present in the group (*meaning that the row is only present in one source and not in the other*)
    - The :class:`data_compare.src.models.RowGroupDifference` object will be returned when there are multiple sources present in the group (*meaning that the row is present in multiple sources and the differences between the rows of the different sources are returned*).

    Example:
        ```python
        import polars as pl
        from data_compare.src.comparator import _get_row_differences_paired

        df0 = pl.DataFrame({"a": [1, 2, 3], "b": [1, 2, 3]})
        df1 = pl.DataFrame({"a": [1, 2, 3], "b": [1, 2, 10]})
        df0_name = "df0"
        df1_name = "df1"

        same_columns, different_columns, row_differences = _get_row_differences_paired(
            df0, df1, df0_name, df1_name, ["a"]
        )
        print(row_differences)
        ```
        Output:
        ```
        [RowGroupDifference(sources=['df0', 'df1'],
        row={'a': [3, 3], 'b': [3, 10]},
        number_of_occurrences=2,
        grouping_columns=['a'],
        column_differences=['b'],
        consise_information={'a': [3, 3], 'b': [3, 10], 'source': ['df0', 'df1']},
        row_with_source={'a': [3, 3], 'b': [3, 10], 'source': ['df0', 'df1']})]
        ```

    Raises:
        ValueError: If the dataframe is not grouped by the grouping columns.

    Args:
        data (pl.DataFrame): The dataframe to compare.
        grouping_columns (list[str]): The columns to group by.

    Returns:
        list[Union[:class:`data_compare.src.models.RowDifference`, :class:`data_compare.src.models.RowGroupDifference`]]: The differences between the rows of the different sources.


    """
    if data.select(grouping_columns).unique().shape[0] != 1:
        raise ValueError("The dataframe must be grouped by the grouping columns.")

    sources: list[str] = list(data["source"].unique())

    if len(sources) == 1:
        row_difference_information: RowDifference = RowDifference(
            source=sources[0],
            row=data.select(sorted(data.columns))
            .drop(["source"])
            .sort("*")
            .to_dict(as_series=False),
            number_of_occurrences=len(data),
            difference_type=RowDifferenceType.MISSING_ROW,
        )
        return row_difference_information

    different_columns: list[str] = []
    to_check_columns: set[str] = (
        set(data.columns) - set(grouping_columns) - {"hash", "source"}
    )
    for col in to_check_columns:
        if len(data[col].unique()) == 1:
            continue

        different_columns.append(col)

    row_grouping_difference: RowGroupDifference = RowGroupDifference(
        sources=sorted(sources),
        row=data.select(sorted(data.columns))
        .drop(["source"])
        .sort("*")
        .to_dict(as_series=False),
        number_of_occurrences=len(data),
        grouping_columns=sorted(grouping_columns),
        column_differences=sorted(different_columns),
        consise_information=data.select(
            sorted(grouping_columns + different_columns + ["source"])
        )
        .sort("*")
        .to_dict(as_series=False),
        row_with_source=data.select(sorted(data.columns))
        .sort("*")
        .to_dict(as_series=False),
    )
    return row_grouping_difference


@convert_to_polars
@check_inputs
def get_row_differences_paired(
    df0: pl.DataFrame,
    df1: pl.DataFrame,
    df0_name: str,
    df_1_name: str,
    grouping_columns: list[str],
) -> tuple[
    list[str], list[ColumnDifference], list[Union[RowDifference, RowGroupDifference]]
]:
    """
    Compares the rows of a dataframe grouped by the grouping columns.

    If you have two dataframes where some columns (or multiple of them) are identifying a row,
    with this function you can find the columns in which they differ, but also have a functionality that
    tells you that there are some missing rows in one of the dataframes.

    Example:
        ```python
        import polars as pl
        from data_compare.src.comparator import get_row_differences_paired
        df0 = pl.DataFrame({"a": [1, 2, 3], "b": [1, 2, 3]})
        df1 = pl.DataFrame({"a": [1, 2, 3], "b": [1, 2, 10]})
        df0_name = "df0"
        df1_name = "df1"
        same_columns, different_columns, row_differences = get_row_differences_paired(
            df0, df1, df0_name, df1_name, ["a"]
        )
        print(row_differences)
        ```
        Output:
        ```
        [RowGroupDifference(sources=['df0', 'df1'],
        row={'a': [3, 3], 'b': [3, 10]},
        number_of_occurrences=2,
        grouping_columns=['a'],
        column_differences=['b'],
        consise_information={'a': [3, 3], 'b': [3, 10], 'source': ['df0', 'df1']},
        row_with_source={'a': [3, 3], 'b': [3, 10], 'source': ['df0', 'df1']})]
        ```

    Raises:
        ValueError: If the pairing columns are not the present in both dataframes.

    Args:
        df0 (pl.DataFrame): The first dataframe.
        df1 (pl.DataFrame): The second dataframe.
        df0_name (str): The name of the first dataframe.
        df_1_name (str): The name of the second dataframe.
        grouping_columns (list[str]): The columns to group by.

    Returns:
        list[str]: The same columns

        list[:class:`data_compare.src.models.ColumnDifference`]: The column differences

        list[Union[:class:`data_compare.src.models.RowDifference`, :class:`data_compare.src.models.RowGroupDifference`]]: The row differences
    """
    same_columns, column_differences, row_differences = get_row_differences(
        df0, df1, df0_name, df_1_name
    )

    if len(set(grouping_columns).difference(same_columns)) > 0:
        raise ValueError(
            "Pairing columns must be the same in both dataframes. "
            f"Pairing columns: {grouping_columns}. Same columns: {same_columns}"
        )

    difference_dataframe: pl.DataFrame = convert_row_differences_to_pandas(
        row_differences
    )
    if len(difference_dataframe) == 0:
        return same_columns, column_differences, row_differences

    row_differences: list[Union[RowDifference, RowGroupDifference]] = []
    for name, dat in difference_dataframe.group_by(grouping_columns):
        difference: Union[RowDifference, RowGroupDifference] = (
            compare_group_column_by_column(dat, grouping_columns)
        )
        row_differences.append(difference)
    return same_columns, column_differences, row_differences


@convert_to_polars
@check_inputs
def get_data_report(
    df0: pl.DataFrame,
    df1: pl.DataFrame,
    df0_name: str,
    df1_name: str,
    grouping_columns: Optional[list[str]] = None,
) -> DataReport:
    """
    Get a data report comparing two dataframes.

    This function returns multiple informations like:
    - the length of the dataframes
    - the columns that are the same in both dataframes (have same column name and data type)
    - the columns that are different in both dataframes (they have different data types or not in both dataframes)
    - the rows that are different in both dataframes (they are not existing or they have different column values if grouping parameter is used)
    - total number of row differences (on columns that are same in both dataframes)
    - total number of row differences in the first dataframe
    - total number of row differences in the second dataframe
    - the ratio of row differences in the first dataframe
    - the ratio of row differences in the second dataframe

    .. note::
       The ratio of row differences is calculated as the
       number of row differences from some of the dataframes divided by the total number of row differences.

    Example:
        ```python
        import polars as pl
        from data_compare.src.comparator import get_data_report
        df0 = pl.DataFrame({"a": [1, 2, 3, 3, 3, 4], "b": [1, 2, 3, 10, 10, 15]})
        df1 = pl.DataFrame({"a": [1, 2, 3, 3, 4, 5], "b": [1, 2, 3, 10, 20, 24]})
        df0_name = "df0"
        df1_name = "df1"
        report = get_data_report(df0, df1, df0_name, df1_name, ["a"])
        print(report.model_dump_json(indent=4))
        ```

        Output:
        ```
        {
            "df0_length": 6,
            "df1_length": 6,
            "df0_name": "df0",
            "df1_name": "df1",
            "comparable_columns": [
                "b",
                "a"
            ],
            "column_differences": [],
            "row_differences": [
                {
                    "source": "df1",
                    "row": {
                        "a": [
                            5
                        ],
                        "b": [
                            24
                        ]
                    },
                    "number_of_occurrences": 1,
                    "difference_type": "MISSING_ROW",
                    "more_information": null
                },
                {
                    "source": "df0",
                    "row": {
                        "a": [
                            3
                        ],
                        "b": [
                            10
                        ]
                    },
                    "number_of_occurrences": 1,
                    "difference_type": "MISSING_ROW",
                    "more_information": null
                },
                {
                    "sources": [
                        "df0",
                        "df1"
                    ],
                    "row": {
                        "a": [
                            4,
                            4
                        ],
                        "b": [
                            15,
                            20
                        ]
                    },
                    "number_of_occurrences": 2,
                    "grouping_columns": [
                        "a"
                    ],
                    "column_differences": [
                        "b"
                    ],
                    "consise_information": {
                        "a": [
                            4,
                            4
                        ],
                        "b": [
                            15,
                            20
                        ],
                        "source": [
                            "df0",
                            "df1"
                        ]
                    },
                    "row_with_source": {
                        "a": [
                            4,
                            4
                        ],
                        "b": [
                            15,
                            20
                        ],
                        "source": [
                            "df0",
                            "df1"
                        ]
                    }
                }
            ],
            "number_of_row_differences": 4,
            "number_of_differences_source_0": 2,
            "number_of_differences_source_1": 2,
            "ratio_of_difference_from_source_0": 0.5,
            "ratio_of_difference_from_source_1": 0.5
        }
        ```

    Args:
        df0 (pl.DataFrame): The first dataframe.
        df1 (pl.DataFrame): The second dataframe.

    Returns:
        :class:`data_compare.src.models.DataReport`: A data report comparing the two dataframes.
    """
    if grouping_columns is None:
        same_columns, column_differences, row_differences = get_row_differences(
            df0, df1, df0_name, df1_name
        )
    else:
        same_columns, column_differences, row_differences = get_row_differences_paired(
            df0, df1, df0_name, df1_name, grouping_columns
        )
    return DataReport(
        df0_length=len(df0),
        df1_length=len(df1),
        df1=df1,
        df0_name=df0_name,
        df1_name=df1_name,
        comparable_columns=same_columns,
        row_differences=row_differences,
        column_differences=column_differences,
    )



from typing import Any, Optional, Union

from pydantic import BaseModel, computed_field
import polars as pl

from data_fingerprint.src.difference_types import (
    ColumnNameDifferenceType,
    ColumnDataTypeDifferenceType,
    RowDifferenceType,
)


class RowDifference(BaseModel):
    """
    Model for row differences.
    """

    source: str
    """The data source name of the row."""

    row: dict[str, Any]
    """The **RAW** row that is different."""

    number_of_occurrences: int
    """The number of times this row is present in the source."""

    difference_type: RowDifferenceType
    """The type of difference."""

    more_information: Optional[Any] = None
    """More information about the difference."""

    def __hash__(self):
        return hash(
            (
                self.source,
                str(self.row),
                self.number_of_occurrences,
                self.difference_type,
                str(self.more_information),
            )
        )


class RowGroupDifference(BaseModel):
    """
    Model for row group differences.
    """

    sources: list[str]
    """The sources of the different rows."""

    row: dict[str, Any]
    """The **RAW** row that is different."""

    number_of_occurrences: int
    """The number of times this difference is present in all sources (**total**)."""

    grouping_columns: list[str]
    """The columns used to group the rows."""

    column_differences: list[str]
    """The columns that are different."""

    consise_information: dict[str, Any]
    """
    More information about the difference. 
    This view only shows the differences between the sources, 
    it is not showing the columns that are the same.
    """

    row_with_source: dict[str, Any]
    """
    The row with the source name as column.
    """

    def __hash__(self):
        return hash(
            (
                str(sorted(self.sources)),
                str(self.row),
                self.number_of_occurrences,
                str(sorted(self.grouping_columns)),
                str(sorted(self.column_differences)),
                str(self.consise_information),
            )
        )


class ColumnDifference(BaseModel):
    """
    Model for column differences.
    """

    source: str
    """The source of the column difference."""

    column_name: str
    """The name of the column that is different."""

    difference_type: Union[ColumnNameDifferenceType, ColumnDataTypeDifferenceType]
    """The type of difference."""

    more_information: Optional[Any] = None
    """More information about the difference."""

    def __hash__(self):
        return hash(
            (
                self.source,
                self.column_name,
                self.difference_type,
                str(self.more_information),
            )
        )


class DataReport(BaseModel):
    """
    Model for data report.
    """

    df0_length: int
    """The length of the first dataframe."""

    df1_length: int
    """The length of the second dataframe."""

    df0_name: str
    """The name of the first dataframe."""

    df1_name: str
    """The name of the second dataframe."""

    comparable_columns: list[str]
    """The columns that are comparable (*same name and same data type*)."""

    column_differences: list[ColumnDifference]
    """The column differences."""

    row_differences: list[Union[RowDifference, RowGroupDifference]]
    """The row differences."""

    @computed_field
    @property
    def number_of_row_differences(self) -> int:
        """
        The number of row differences.
        """
        return sum([rd.number_of_occurrences for rd in self.row_differences])

    @computed_field
    @property
    def number_of_differences_source_0(self) -> int:
        """
        The number of row differences from the first dataframe.
        """
        counter: int = 0
        for rd in self.row_differences:
            if isinstance(rd, RowDifference) and rd.source == self.df1_name:
                continue

            if isinstance(rd, RowDifference) and rd.source == self.df0_name:
                counter += rd.number_of_occurrences
                continue

            counter += sum(
                [1 for x in rd.consise_information["source"] if x == self.df0_name]
            )

        return counter

    @computed_field
    @property
    def number_of_differences_source_1(self) -> int:
        """
        The number of row differences from the second dataframe.
        """
        counter: int = 0
        for rd in self.row_differences:
            if isinstance(rd, RowDifference) and rd.source == self.df0_name:
                continue

            if isinstance(rd, RowDifference) and rd.source == self.df1_name:
                counter += rd.number_of_occurrences
                continue

            counter += sum(
                [1 for x in rd.consise_information["source"] if x == self.df1_name]
            )

        return counter

    @computed_field
    @property
    def ratio_of_difference_from_source_0(self) -> float:
        """
        The ratio of row differences from the first dataframe.
        This is the number of row differences from the first dataframe divided by the total number of row differences.

        If there are no row differences, this will return 0.0.
        """
        if self.number_of_row_differences == 0:
            return 0.0

        return self.number_of_differences_source_0 / self.number_of_row_differences

    @computed_field
    @property
    def ratio_of_difference_from_source_1(self) -> float:
        """
        The ratio of row differences from the second dataframe.
        This is the number of row differences from the second dataframe divided by the total number of row differences.

        If there are no row differences, this will return 0.0.
        """
        if self.number_of_row_differences == 0:
            return 0.0

        return self.number_of_differences_source_1 / self.number_of_row_differences



from typing import Any, Callable
import json

import pandas as pd
import polars as pl


def _raise_same_column_names(argument: Any, **kwargs) -> None:
    """
    Check if the argument is a `pandas.DataFrame` and has more then one occurance of the same column name.
    If so, raise a ValueError.

    > Note: The interesting thing is that pandas allows to have multiple columns with the same name.

    Parameters:
        argument (Any): The argument to check.

    Raises:
        ValueError: If the argument is a `pandas.DataFrame` and has more then one occurance of the same column name.

    Returns:
        None
    """
    if not isinstance(argument, pd.DataFrame):
        return

    df: pd.DataFrame = argument
    column_names_list: list[str] = list(df.columns)
    column_names_set: set[str] = set(column_names_list)

    if len(column_names_list) == len(column_names_set):
        return

    columns_with_same_name: dict[str, int] = {
        column_name: column_names_list.count(column_name)
        for column_name in column_names_set
    }
    raise ValueError(
        "Column names are not unique. Distribution of column names: "
        f"{json.dumps(columns_with_same_name, indent=4)}"
    )


def _raise_hash_column_name(argument: Any, **kwargs) -> None:
    """
    Check if the argument is a `pandas.DataFrame` or `polars.DataFrame` and has a column named `hash`.
    If so, raise a `ValueError`.

    Parameters:
        argument (Any): The argument to check.

    Raises:
        ValueError: If the argument is a `DataFrame` and has a column named `hash`.

    Returns:
        None
    """
    if not isinstance(argument, pd.DataFrame) and not isinstance(
        argument, pl.DataFrame
    ):
        return

    column_names_list: list[str] = list(argument.columns)
    column_names_set: set[str] = set(column_names_list)

    if "hash" in column_names_set:
        raise ValueError("Column names cannot contain 'hash'")


def _raise_source_column_name(argument: Any, **kwargs) -> None:
    """
    Check if the argument is a `pandas.DataFrame` or `polars.DataFrame` and has a column named `source`.
    If so, raise a `ValueError`.

    Parameters:
        argument (Any): The argument to check.

    Raises:
        ValueError: If the argument is a `DataFrame` and has a column named `source`.

    Returns:
        None
    """
    if not isinstance(argument, pd.DataFrame) and not isinstance(
        argument, pl.DataFrame
    ):
        return

    column_names_list: list[str] = list(argument.columns)
    column_names_set: set[str] = set(column_names_list)
    if "source" in column_names_set:
        raise ValueError("Column names cannot contain 'source'")


def _raise_source_names(argument: Any, source_names: list[str]) -> None:
    """
    Check if the argument is a string (*meaning it is a source name*) and if it is already in the list of source names.
    If so, raise a `ValueError`.

    Parameters:
        argument (Any): The argument to check.
        source_names (list[str]): The list of already existing source names.

    Raises:
        ValueError: If the argument is a string and is already in the list of source names.

    Returns:
        None
    """
    if not isinstance(argument, str):
        return

    if argument in source_names:
        raise ValueError(f"Source name already exists: {argument}")

    source_names.append(argument)


_rules_for_inputs: list[Callable[[Any], None]] = [
    _raise_same_column_names,
    _raise_hash_column_name,
    _raise_source_column_name,
    _raise_source_names,
]
"""Rules for checking the inputs, they are applied in the order they are defined."""


def check_inputs(func) -> None:
    """
    Decorator for checking the input arguments against a set of input rules.
    It is developed to be used for functions that take dataframes and source names as inputs.

    Something like this:

    ```python
    @check_inputs
    def func(
        source_0: pl.DataFrame,
        source_1: pl.DataFrame,
        source_0_name: str,
        source_1_name: str,
    ) -> None:
        pass
    ```
    If you look at the :mod:`data_compare.src.comparator` you will see that this decorator is used there on
    all the functions that take dataframes and source names as inputs.

    If any rule is violated, an exception will be raised.
    The rules are defined in the _rules_for_inputs list.

    .. note::
        The rules are applied in the order they are defined. There are rules:
        - there must not be any duplicate column names in the dataframes.
        - there must not be any column named "hash" in the dataframes
        - there must not be any column named "source" in the dataframes
        - there must not be any duplicate source names

    Parameters:
        *args (Any): The inputs to check.

    Raises:
        Exception: If any rule is violated.

    Returns:
        None
    """

    def wrapper(*args, **kwargs):
        source_names: list[str] = []
        for arg in args:
            for rule in _rules_for_inputs:
                rule(arg, source_names=source_names)

        for value in kwargs.values():
            for rule in _rules_for_inputs:
                rule(value, source_names=source_names)
        return func(*args, **kwargs)

    return wrapper
